@@ -25,8 +25,6 @@
 #include "pool-buffer.h"
 #include "seat.h"
 #include "swaylock.h"
-#include "wlr-input-inhibitor-unstable-v1-client-protocol.h"
-#include "wlr-layer-shell-unstable-v1-client-protocol.h"
 #include "ext-session-lock-v1-client-protocol.h"
 
 static uint32_t parse_color(const char *color) {
@@ -96,24 +94,24 @@ static void daemonize(void) {
 
 static void destroy_surface(struct swaylock_surface *surface) {
 	wl_list_remove(&surface->link);
-	if (surface->layer_surface != NULL) {
-		zwlr_layer_surface_v1_destroy(surface->layer_surface);
-	}
 	if (surface->ext_session_lock_surface_v1 != NULL) {
 		ext_session_lock_surface_v1_destroy(surface->ext_session_lock_surface_v1);
 	}
+	if (surface->subsurface) {
+		wl_subsurface_destroy(surface->subsurface);
+	}
+	if (surface->child) {
+		wl_surface_destroy(surface->child);
+	}
 	if (surface->surface != NULL) {
 		wl_surface_destroy(surface->surface);
 	}
-	destroy_buffer(&surface->buffers[0]);
-	destroy_buffer(&surface->buffers[1]);
 	destroy_buffer(&surface->indicator_buffers[0]);
 	destroy_buffer(&surface->indicator_buffers[1]);
-	wl_output_destroy(surface->output);
+	wl_output_release(surface->output);
 	free(surface);
 }
 
-static const struct zwlr_layer_surface_v1_listener layer_surface_listener;
 static const struct ext_session_lock_surface_v1_listener ext_session_lock_surface_v1_listener;
 
 static cairo_surface_t *select_image(struct swaylock_state *state,
@@ -140,28 +138,10 @@ static void create_surface(struct swaylock_surface *surface) {
 	assert(surface->subsurface);
 	wl_subsurface_set_sync(surface->subsurface);
 
-	if (state->ext_session_lock_v1) {
-		surface->ext_session_lock_surface_v1 = ext_session_lock_v1_get_lock_surface(
-				state->ext_session_lock_v1, surface->surface, surface->output);
-		ext_session_lock_surface_v1_add_listener(surface->ext_session_lock_surface_v1,
-				&ext_session_lock_surface_v1_listener, surface);
-	} else {
-		surface->layer_surface = zwlr_layer_shell_v1_get_layer_surface(
-				state->layer_shell, surface->surface, surface->output,
-				ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY, "lockscreen");
-
-		zwlr_layer_surface_v1_set_size(surface->layer_surface, 0, 0);
-		zwlr_layer_surface_v1_set_anchor(surface->layer_surface,
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT);
-		zwlr_layer_surface_v1_set_exclusive_zone(surface->layer_surface, -1);
-		zwlr_layer_surface_v1_set_keyboard_interactivity(
-				surface->layer_surface, true);
-		zwlr_layer_surface_v1_add_listener(surface->layer_surface,
-				&layer_surface_listener, surface);
-	}
+	surface->ext_session_lock_surface_v1 = ext_session_lock_v1_get_lock_surface(
+		state->ext_session_lock_v1, surface->surface, surface->output);
+	ext_session_lock_surface_v1_add_listener(surface->ext_session_lock_surface_v1,
+		&ext_session_lock_surface_v1_listener, surface);
 
 	if (surface_is_opaque(surface) &&
 			surface->state->args.mode != BACKGROUND_MODE_CENTER &&
@@ -173,33 +153,9 @@ static void create_surface(struct swaylock_surface *surface) {
 		wl_region_destroy(region);
 	}
 
-	if (!state->ext_session_lock_v1) {
-		wl_surface_commit(surface->surface);
-	}
+	surface->created = true;
 }
 
-static void layer_surface_configure(void *data,
-		struct zwlr_layer_surface_v1 *layer_surface,
-		uint32_t serial, uint32_t width, uint32_t height) {
-	struct swaylock_surface *surface = data;
-	surface->width = width;
-	surface->height = height;
-	zwlr_layer_surface_v1_ack_configure(layer_surface, serial);
-	render_frame_background(surface);
-	render_frame(surface);
-}
-
-static void layer_surface_closed(void *data,
-		struct zwlr_layer_surface_v1 *layer_surface) {
-	struct swaylock_surface *surface = data;
-	destroy_surface(surface);
-}
-
-static const struct zwlr_layer_surface_v1_listener layer_surface_listener = {
-	.configure = layer_surface_configure,
-	.closed = layer_surface_closed,
-};
-
 static void ext_session_lock_surface_v1_handle_configure(void *data,
 		struct ext_session_lock_surface_v1 *lock_surface, uint32_t serial,
 		uint32_t width, uint32_t height) {
@@ -280,7 +236,10 @@ static void handle_wl_output_mode(void *data, struct wl_output *output,
 }
 
 static void handle_wl_output_done(void *data, struct wl_output *output) {
-	// Who cares
+	struct swaylock_surface *surface = data;
+	if (!surface->created && surface->state->run_display) {
+		create_surface(surface);
+	}
 }
 
 static void handle_wl_output_scale(void *data, struct wl_output *output,
@@ -347,12 +306,6 @@ static void handle_global(void *data, struct wl_registry *registry,
 			calloc(1, sizeof(struct swaylock_seat));
 		swaylock_seat->state = state;
 		wl_seat_add_listener(seat, &seat_listener, swaylock_seat);
-	} else if (strcmp(interface, zwlr_layer_shell_v1_interface.name) == 0) {
-		state->layer_shell = wl_registry_bind(
-				registry, name, &zwlr_layer_shell_v1_interface, 1);
-	} else if (strcmp(interface, zwlr_input_inhibit_manager_v1_interface.name) == 0) {
-		state->input_inhibit_manager = wl_registry_bind(
-				registry, name, &zwlr_input_inhibit_manager_v1_interface, 1);
 	} else if (strcmp(interface, wl_output_interface.name) == 0) {
 		struct swaylock_surface *surface =
 			calloc(1, sizeof(struct swaylock_surface));
@@ -362,11 +315,6 @@ static void handle_global(void *data, struct wl_registry *registry,
 		surface->output_global_name = name;
 		wl_output_add_listener(surface->output, &_wl_output_listener, surface);
 		wl_list_insert(&state->surfaces, &surface->link);
-
-		if (state->run_display) {
-			create_surface(surface);
-			wl_display_roundtrip(state->display);
-		}
 	} else if (strcmp(interface, ext_session_lock_manager_v1_interface.name) == 0) {
 		state->ext_session_lock_manager_v1 = wl_registry_bind(registry, name,
 				&ext_session_lock_manager_v1_interface, 1);
@@ -579,6 +527,7 @@ static int parse_options(int argc, char **argv, struct swaylock_state *state,
 		{"debug", no_argument, NULL, 'd'},
 		{"ignore-empty-password", no_argument, NULL, 'e'},
 		{"daemonize", no_argument, NULL, 'f'},
+		{"ready-fd", required_argument, NULL, 'R'},
 		{"help", no_argument, NULL, 'h'},
 		{"image", required_argument, NULL, 'i'},
 		{"disable-caps-lock-text", no_argument, NULL, 'L'},
@@ -645,6 +594,8 @@ static int parse_options(int argc, char **argv, struct swaylock_state *state,
 			"Show current count of failed authentication attempts.\n"
 		"  -f, --daemonize                  "
 			"Detach from the controlling terminal after locking.\n"
+		"  -R, --ready-fd <fd>              "
+			"File descriptor to send readiness notifications to.\n"
 		"  -h, --help                       "
 			"Show help message and quit.\n"
 		"  -i, --image [[<output>]:]<path>  "
@@ -754,7 +705,7 @@ static int parse_options(int argc, char **argv, struct swaylock_state *state,
 	optind = 1;
 	while (1) {
 		int opt_idx = 0;
-		c = getopt_long(argc, argv, "c:deFfhi:kKLlnrs:tuvC:", long_options,
+		c = getopt_long(argc, argv, "c:deFfhi:kKLlnrs:tuvC:R:", long_options,
 				&opt_idx);
 		if (c == -1) {
 			break;
@@ -788,6 +739,11 @@ static int parse_options(int argc, char **argv, struct swaylock_state *state,
 				state->args.daemonize = true;
 			}
 			break;
+		case 'R':
+			if (state) {
+				state->args.ready_fd = strtol(optarg, NULL, 10);
+			}
+			break;
 		case 'i':
 			if (state) {
 				load_image(optarg, state);
@@ -1122,7 +1078,7 @@ static void comm_in(int fd, short mask, void *data) {
 		state.run_display = false;
 	} else {
 		state.auth_state = AUTH_STATE_INVALID;
-		schedule_indicator_clear(&state);
+		schedule_auth_idle(&state);
 		++state.failed_attempts;
 		damage_state(&state);
 	}
@@ -1182,7 +1138,8 @@ int main(int argc, char **argv) {
 		.show_keyboard_layout = false,
 		.hide_keyboard_layout = false,
 		.show_failed_attempts = false,
-		.indicator_idle_visible = false
+		.indicator_idle_visible = false,
+		.ready_fd = -1,
 	};
 	wl_list_init(&state.images);
 	set_default_colors(&state.args.colors);
@@ -1231,7 +1188,11 @@ int main(int argc, char **argv) {
 
 	if (pipe(sigusr_fds) != 0) {
 		swaylock_log(LOG_ERROR, "Failed to pipe");
-		return 1;
+		return EXIT_FAILURE;
+	}
+	if (fcntl(sigusr_fds[1], F_SETFL, O_NONBLOCK) == -1) {
+		swaylock_log(LOG_ERROR, "Failed to make pipe end nonblocking");
+		return EXIT_FAILURE;
 	}
 
 	wl_list_init(&state.surfaces);
@@ -1244,10 +1205,14 @@ int main(int argc, char **argv) {
 				"WAYLAND_DISPLAY environment variable.");
 		return EXIT_FAILURE;
 	}
+	state.eventloop = loop_create();
 
 	struct wl_registry *registry = wl_display_get_registry(state.display);
 	wl_registry_add_listener(registry, &registry_listener, &state);
-	wl_display_roundtrip(state.display);
+	if (wl_display_roundtrip(state.display) == -1) {
+		swaylock_log(LOG_ERROR, "wl_display_roundtrip() failed");
+		return EXIT_FAILURE;
+	}
 
 	if (!state.compositor) {
 		swaylock_log(LOG_ERROR, "Missing wl_compositor");
@@ -1264,27 +1229,17 @@ int main(int argc, char **argv) {
 		return 1;
 	}
 
-	if (state.ext_session_lock_manager_v1) {
-		swaylock_log(LOG_DEBUG, "Using ext-session-lock-v1");
-		state.ext_session_lock_v1 = ext_session_lock_manager_v1_lock(state.ext_session_lock_manager_v1);
-		ext_session_lock_v1_add_listener(state.ext_session_lock_v1,
-				&ext_session_lock_v1_listener, &state);
-	} else if (state.layer_shell && state.input_inhibit_manager) {
-		swaylock_log(LOG_DEBUG, "Using wlr-layer-shell + wlr-input-inhibitor");
-		zwlr_input_inhibit_manager_v1_get_inhibitor(state.input_inhibit_manager);
-	} else {
-		swaylock_log(LOG_ERROR, "Missing ext-session-lock-v1, wlr-layer-shell "
-				"and wlr-input-inhibitor");
+	if (!state.ext_session_lock_manager_v1) {
+		swaylock_log(LOG_ERROR, "Missing ext-session-lock-v1");
 		return 1;
 	}
 
+	state.ext_session_lock_v1 = ext_session_lock_manager_v1_lock(state.ext_session_lock_manager_v1);
+	ext_session_lock_v1_add_listener(state.ext_session_lock_v1,
+		&ext_session_lock_v1_listener, &state);
+
 	if (wl_display_roundtrip(state.display) == -1) {
 		free(state.args.font);
-		if (state.input_inhibit_manager) {
-			swaylock_log(LOG_ERROR, "Exiting - failed to inhibit input:"
-					" is another lockscreen already running?");
-			return 2;
-		}
 		return 1;
 	}
 
@@ -1296,33 +1251,37 @@ int main(int argc, char **argv) {
 		create_surface(surface);
 	}
 
-	if (state.ext_session_lock_manager_v1) {
-		while (!state.locked) {
-			if (wl_display_dispatch(state.display) < 0) {
-				swaylock_log(LOG_ERROR, "wl_display_dispatch() failed");
-				return 2;
-			}
-		}
-	} else {
-		if (wl_display_roundtrip(state.display) < 0) {
-			swaylock_log(LOG_ERROR, "wl_display_roundtrip() failed");
+	while (!state.locked) {
+		if (wl_display_dispatch(state.display) < 0) {
+			swaylock_log(LOG_ERROR, "wl_display_dispatch() failed");
 			return 2;
 		}
-		state.locked = true;
 	}
 
+	if (state.args.ready_fd >= 0) {
+		if (write(state.args.ready_fd, "\n", 1) != 1) {
+			swaylock_log(LOG_ERROR, "Failed to send readiness notification");
+			return 2;
+		}
+		close(state.args.ready_fd);
+		state.args.ready_fd = -1;
+	}
 	if (state.args.daemonize) {
 		daemonize();
 	}
 
-	state.eventloop = loop_create();
 	loop_add_fd(state.eventloop, wl_display_get_fd(state.display), POLLIN,
 			display_in, NULL);
 
 	loop_add_fd(state.eventloop, get_comm_reply_fd(), POLLIN, comm_in, NULL);
 
 	loop_add_fd(state.eventloop, sigusr_fds[0], POLLIN, term_in, NULL);
-	signal(SIGUSR1, do_sigusr);
+
+	struct sigaction sa;
+	sa.sa_handler = do_sigusr;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART;
+	sigaction(SIGUSR1, &sa, NULL);
 
 	state.run_display = true;
 	while (state.run_display) {
@@ -1333,10 +1292,8 @@ int main(int argc, char **argv) {
 		loop_poll(state.eventloop);
 	}
 
-	if (state.ext_session_lock_v1) {
-		ext_session_lock_v1_unlock_and_destroy(state.ext_session_lock_v1);
-		wl_display_roundtrip(state.display);
-	}
+	ext_session_lock_v1_unlock_and_destroy(state.ext_session_lock_v1);
+	wl_display_roundtrip(state.display);
 
 	free(state.args.font);
 	cairo_destroy(state.test_cairo);
