@@ -46,28 +46,50 @@ static void append_ch(struct swaylock_password *pw, uint32_t codepoint) {
 	pw->len += utf8_size;
 }
 
-static void clear_indicator(void *data) {
+static void set_input_idle(void *data) {
 	struct swaylock_state *state = data;
-	state->clear_indicator_timer = NULL;
+	state->input_idle_timer = NULL;
+	state->input_state = INPUT_STATE_IDLE;
+	damage_state(state);
+}
+
+static void set_auth_idle(void *data) {
+	struct swaylock_state *state = data;
+	state->auth_idle_timer = NULL;
 	state->auth_state = AUTH_STATE_IDLE;
 	damage_state(state);
 }
 
-void schedule_indicator_clear(struct swaylock_state *state) {
-	if (state->clear_indicator_timer) {
-		loop_remove_timer(state->eventloop, state->clear_indicator_timer);
+static void schedule_input_idle(struct swaylock_state *state) {
+	if (state->input_idle_timer) {
+		loop_remove_timer(state->eventloop, state->input_idle_timer);
 	}
-	state->clear_indicator_timer = loop_add_timer(
-			state->eventloop, 3000, clear_indicator, state);
+	state->input_idle_timer = loop_add_timer(
+		state->eventloop, 1500, set_input_idle, state);
+}
+
+static void cancel_input_idle(struct swaylock_state *state) {
+	if (state->input_idle_timer) {
+		loop_remove_timer(state->eventloop, state->input_idle_timer);
+		state->input_idle_timer = NULL;
+	}
+}
+
+void schedule_auth_idle(struct swaylock_state *state) {
+	if (state->auth_idle_timer) {
+		loop_remove_timer(state->eventloop, state->auth_idle_timer);
+	}
+	state->auth_idle_timer = loop_add_timer(
+		state->eventloop, 3000, set_auth_idle, state);
 }
 
 static void clear_password(void *data) {
 	struct swaylock_state *state = data;
 	state->clear_password_timer = NULL;
-	state->auth_state = AUTH_STATE_CLEAR;
+	state->input_state = INPUT_STATE_CLEAR;
+	schedule_input_idle(state);
 	clear_password_buffer(&state->password);
 	damage_state(state);
-	schedule_indicator_clear(state);
 }
 
 static void schedule_password_clear(struct swaylock_state *state) {
@@ -78,27 +100,39 @@ static void schedule_password_clear(struct swaylock_state *state) {
 			state->eventloop, 10000, clear_password, state);
 }
 
+static void cancel_password_clear(struct swaylock_state *state) {
+	if (state->clear_password_timer) {
+		loop_remove_timer(state->eventloop, state->clear_password_timer);
+		state->clear_password_timer = NULL;
+	}
+}
+
 static void submit_password(struct swaylock_state *state) {
 	if (state->args.ignore_empty && state->password.len == 0) {
 		return;
 	}
 
+	state->input_state = INPUT_STATE_IDLE;
 	state->auth_state = AUTH_STATE_VALIDATING;
+	cancel_password_clear(state);
+	cancel_input_idle(state);
 
 	if (!write_comm_request(&state->password)) {
 		state->auth_state = AUTH_STATE_INVALID;
-		schedule_indicator_clear(state);
+		schedule_auth_idle(state);
 	}
 
 	damage_state(state);
 }
 
+static void update_highlight(struct swaylock_state *state) {
+	// Advance a random amount between 1/4 and 3/4 of a full turn
+	state->highlight_start =
+		(state->highlight_start + (rand() % 1024) + 512) % 2048;
+}
+
 void swaylock_handle_key(struct swaylock_state *state,
 		xkb_keysym_t keysym, uint32_t codepoint) {
-	// Ignore input events if validating
-	if (state->auth_state == AUTH_STATE_VALIDATING) {
-		return;
-	}
 
 	switch (keysym) {
 	case XKB_KEY_KP_Enter: /* fallthrough */
@@ -107,20 +141,29 @@ void swaylock_handle_key(struct swaylock_state *state,
 		break;
 	case XKB_KEY_Delete:
 	case XKB_KEY_BackSpace:
-		if (backspace(&state->password)) {
-			state->auth_state = AUTH_STATE_BACKSPACE;
+		if (state->xkb.control) {
+			clear_password_buffer(&state->password);
+			state->input_state = INPUT_STATE_CLEAR;
+			cancel_password_clear(state);
 		} else {
-			state->auth_state = AUTH_STATE_CLEAR;
+			if (backspace(&state->password) && state->password.len != 0) {
+				state->input_state = INPUT_STATE_BACKSPACE;
+				schedule_password_clear(state);
+				update_highlight(state);
+			} else {
+				state->input_state = INPUT_STATE_CLEAR;
+				cancel_password_clear(state);
+			}
 		}
+		schedule_input_idle(state);
 		damage_state(state);
-		schedule_indicator_clear(state);
-		schedule_password_clear(state);
 		break;
 	case XKB_KEY_Escape:
 		clear_password_buffer(&state->password);
-		state->auth_state = AUTH_STATE_CLEAR;
+		state->input_state = INPUT_STATE_CLEAR;
+		cancel_password_clear(state);
+		schedule_input_idle(state);
 		damage_state(state);
-		schedule_indicator_clear(state);
 		break;
 	case XKB_KEY_Caps_Lock:
 	case XKB_KEY_Shift_L:
@@ -133,10 +176,10 @@ void swaylock_handle_key(struct swaylock_state *state,
 	case XKB_KEY_Alt_R:
 	case XKB_KEY_Super_L:
 	case XKB_KEY_Super_R:
-		state->auth_state = AUTH_STATE_INPUT_NOP;
-		damage_state(state);
-		schedule_indicator_clear(state);
+		state->input_state = INPUT_STATE_NEUTRAL;
 		schedule_password_clear(state);
+		schedule_input_idle(state);
+		damage_state(state);
 		break;
 	case XKB_KEY_m: /* fallthrough */
 	case XKB_KEY_d:
@@ -150,19 +193,21 @@ void swaylock_handle_key(struct swaylock_state *state,
 	case XKB_KEY_u:
 		if (state->xkb.control) {
 			clear_password_buffer(&state->password);
-			state->auth_state = AUTH_STATE_CLEAR;
+			state->input_state = INPUT_STATE_CLEAR;
+			cancel_password_clear(state);
+			schedule_input_idle(state);
 			damage_state(state);
-			schedule_indicator_clear(state);
 			break;
 		}
 		// fallthrough
 	default:
 		if (codepoint) {
 			append_ch(&state->password, codepoint);
-			state->auth_state = AUTH_STATE_INPUT;
-			damage_state(state);
-			schedule_indicator_clear(state);
+			state->input_state = INPUT_STATE_LETTER;
 			schedule_password_clear(state);
+			schedule_input_idle(state);
+			update_highlight(state);
+			damage_state(state);
 		}
 		break;
 	}
